<?php

namespace Drupal\search_api_coveo;

use Drupal\Core\Config\ImmutableConfig;
use Drupal\key\Entity\Key;
use Drupal\key\KeyRepositoryInterface;
use Drupal\Core\State\StateInterface;

/**
 * Implements a credential service to support multiple providers.
 */
class CredentialProvider {

  /**
   * Key module service conditionally injected.
   *
   * @var \Drupal\key\KeyRepositoryInterface
   */
  protected $keyRepository;

  /**
   * The Drupal state api.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * KeyService constructor.
   *
   * @param \Drupal\Core\State\StateInterface $state
   *   The key value store to use.
   */
  public function __construct(StateInterface $state) {
    $this->state = $state;
  }

  /**
   * Provides a means to our services.yml file to conditionally inject service.
   *
   * @param \Drupal\key\KeyRepositoryInterface $repository
   *   The injected service, if it exists.
   */
  public function setKeyRepository(KeyRepositoryInterface $repository) {
    $this->keyRepository = $repository;
  }

  /**
   * Detects if key module service was injected.
   *
   * @return bool
   *   True if the KeyRepository is present.
   */
  public function additionalProviders() {
    return $this->keyRepository instanceof KeyRepositoryInterface;
  }

  /**
   * Get the provided credentials.
   *
   * @param array $config
   *   An array of search_api_coveo plugin configuration.
   *
   * @return array|string
   *   The value of the configured key.
   */
  public function getCredentials(array $config) {
    $credentials = [];
    $provider = $this->getProvider($config);
    if (empty($provider)) {
      return $credentials;
    }
    switch ($provider) {
      case 'key':
        $keyEntity = $this->keyRepository->getKey($this->getStorageKey($config));
        if ($keyEntity instanceof Key) {
          // A key was found in the repository.
          $credentials = $keyEntity->getKeyValues();
        }
        break;

      default:
        $credentials = $this->state->get($this->getStorageKey($config));
    }

    return $credentials;
  }

  /**
   * Helper function to get the string used as a storage key by the provider.
   * @param array $config
   *
   * @return string
   */
  protected function getStorageKey(array $config) {
    return $config['storage_key'] ?? '';
  }

  /**
   * Helper function to get string identifying the credential provider.
   *
   * @param array $config
   *
   * @return string
   */
  protected function getProvider(array $config) {
    return $config['credential_provider'] ?? '';
  }

}
