<?php

namespace Drupal\search_api_coveo;

use Coveo\Search\SDK\SDKPushPHP\Document;
use Coveo\Search\SDK\SDKPushPHP\Push;
use Drupal\Component\Serialization\Json;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ServerException;

/**
 * Entry point in the PHP API.
 */
class SearchApiCoveoClient {

  use LoggerTrait;

  /**
   * Search API client context.
   *
   * @var \Drupal\search_api_coveo\SeachApiClientContext
   */
  public $clientContext;

  /**
   * Coveo Push Client.
   *
   * @var \Coveo\Search\SDK\SDKPushPHP\Push
   */
  public $push;


  /**
   * The HTTP Client.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Coveo Search initialization.
   *
   * @param string $sourceID
   *   The push source ID.
   * @param string $organizationID
   *   The organizationID/application ID.
   * @param string $apiKey
   *   Valid API key for the service.
   * @param string $host
   *   The push rest api url.
   * @param \Psr\Log\LoggerInterface $logger
   *   Logger.
   *
   * @throws \Exception
   */
  public function __construct($sourceID, $organizationID, $apiKey, $host, $logger) {
    $this->clientContext = new SeachApiClientContext($organizationID, $sourceID, $apiKey, $host);
    $this->setLogger($logger);

    $this->push = new Push($sourceID, $organizationID, $apiKey, $host, $this->getLogger());
    $this->httpClient = \Drupal::httpClient();
  }

  /**
   * Request query builder.
   *
   * @param array $args
   *   Query arguments.
   *
   * @return string
   *   Query.
   */
  public static function buildQuery(array $args) {
    return '';
  }

  /**
   * Send a request.
   *
   * @param string $url
   *   Url.
   * @param string $method
   *   HTTP method.
   * @param array $params
   *   Request parameters.
   * @param array $data
   *   Data.
   *
   * @return mixed
   *   Request response.
   *
   * @throws \Exception
   *
   * @see https://docs.coveo.com/en/1471/build-a-search-ui/troubleshoot-query-error-codes#403-forbidden
   */
  public function doRequest($url, $method, array $params, array $data) {

    if ($params != NULL && count($params) > 0) {
      $params2 = [];
      foreach ($params as $key => $val) {
        if (is_array($val)) {
          $params2[$key] = Json::encode($val);
        }
        else {
          $params2[$key] = $val;
        }
      }
      $url .= '?' . http_build_query($params2);
    }
    $headers = array_merge($this->clientContext->defaultHeaders, ['Connection' => 'close']);

    $options = [
      'headers' => $headers,
    ];

    if ($method === 'POST' ||  $method === 'PUT') {
      $options['body'] = ($data) ? Json::encode($data) : '{}';
    }

    try {
      $response = $this->httpClient->request($method, $url, $options);
    }
    catch (ClientException $e) {
      return $this->handleErrorMessage($e);
    }
    catch (ServerException $e) {
      return $this->handleErrorMessage($e);
    }
    catch (\Exception $e) {
      $this->logException($e);
      \Drupal::messenger()->addError($this->t("Request failed. Check recent log messages for more details."));
      return FALSE;
    }

    $status = $response->getStatusCode();
    $body = $response->getBody()->getContents();
    // check if response is successfull. in the range 200.
    if (intval($status / 100) == 2) {
      if (in_array($method, ['PUT', 'DELETE']) && (empty($body) || $body == 'null')) {
        return $body;
      }
    }

    $answer = Json::decode($body);
    return $answer;
  }

  /**
   * Start Batch.
   *
   * @param bool $updateStatus
   *   Update Push source status.
   * @param bool $deleteOlder
   *   Delete Items older than.
   */
  public function startBatch(bool $updateStatus = NULL, bool $deleteOlder = NULL) {
    $result = TRUE;
    try {
      $result = $this->push->Start($updateStatus, $deleteOlder);
      if ($result === FALSE) {
        $variables['%start_ordering_id'] = (string) $this->push->StartOrderingId;
        $this->logMessage('Failed to start batch %start_ordering_id.', $variables);
      }
    }
    catch (Exception $ex) {
      $this->logException($ex);
      return FALSE;
    }
    return $result;
  }

  /**
   * Set Size Max Request.
   *
   * @param int $size
   *   Size.
   */
  public function setSizeMaxRequest(int $size) {
    $result = TRUE;
    try {
      $result = $this->push->SetSizeMaxRequest($size);
      if ($result === FALSE) {
        $variables['%size'] = $size;
        $this->logMessage('Failed to set max batch size to %size.', $variables);
      }
    }
    catch (Exception $ex) {
      $this->logException($ex);
      return FALSE;
    }
    return $result;
  }

  /**
   * Add Document to batch.
   *
   * @param \Coveo\Search\SDK\SDKPushPHP\Document $document
   *   The coveo document.
   */
  public function addDocument(Document $document) {
    $result = TRUE;
    try {
      $result = $this->push->Add($document);
      if ($result === FALSE) {
        $variables['%document_id'] = ($document != NULL) ? $document->DocumentId : '<empty document id>';
        $this->logMessage('Add coveo document to batch with id %document_id failed.', $variables);
      }
    }
    catch (\Exception $ex) {
      $this->logException($ex);
      return FALSE;
    }
    return $result;
  }

  /**
   * End batch.
   *
   * @param bool $updateStatus
   *   Update Push source status.
   * @param bool $deleteOlder
   *   Delete items older than.
   */
  public function endBatch(bool $updateStatus = NULL, bool $deleteOlder = NULL) {
    $result = FALSE;
    try {
      $result = $this->push->End($updateStatus, $deleteOlder);
      if ($result === FALSE) {
        $variables['%start_ordering_id'] = (string) $this->push->StartOrderingId;
        $this->logMessage('Failed to end batch %start_ordering_id.', $variables);
      }
    }
    catch (\Exception $ex) {
      $this->logException($ex);
      return FALSE;
    }
    return $result;
  }

  /**
   * This function deletes the index content.
   *
   * Settings and index specific API keys are kept untouched.
   *
   * @return bool
   *   Request response.
   *
   * @throws \Exception
   */
  public function deleteOlderThan() {
    $result = FALSE;
    try {
      $start_ordering_id = $this->push->CreateOrderingId();
      $result = $this->push->DeleteOlderThan($start_ordering_id);
      if ($result === FALSE) {
        $variables['%start_ordering_id'] = $start_ordering_id;
        $this->logMessage('Delete documents older than %start_ordering_id failed.', $variables);
      }
    }
    catch (\Exception $ex) {
      $this->logException($ex);
      return FALSE;
    }
    return $result;
  }

  /**
   * Delete Single Document.
   *
   * @param string $documentId
   *   Coveo Document Id.
   * @param bool $updateStatus
   *   Update the Push source status.
   * @param int $orderingId
   *   Ordering Id.
   * @param bool $deleteChildren
   *   Delete Coveo item children.
   */
  public function deleteItem(string $documentId, bool $updateStatus = NULL, int $orderingId = NULL, bool $deleteChildren = NULL) {
    $result = FALSE;
    try {
      $result = $this->push->RemoveSingleDocument($documentId, $updateStatus, $orderingId, $deleteChildren);
      if ($result === FALSE) {
        $variables['%documentId'] = $documentId;
        $this->logMessage('Remove single document with Id %documentId failed.', $variables);
      }
    }
    catch (\Exception $ex) {
      $this->logException($ex);
      return FALSE;
    }
    return $result;
  }

  /**
   * Get COVEO Organizations.
   *
   * @return array
   *   List of COVEO organizations.
   */
  public function getOrganizations() {
    $url = $this->clientContext->organizationHost;
    return $this->doRequest($url, 'GET', [], []);
  }

  /**
   * Saves a field into coveo.
   *
   * @param array $bulk
   *   An array to bulk create fields.
   *
   * @return mixed
   *   The response.
   *
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/createFieldUsingPOST_2
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/createFieldsUsingPOST_2
   */
  public function createFields(array $bulk = []) {
    if (empty($bulk)) {
      return FALSE;
    }
    $url = $this->clientContext->fieldHost . $this->clientContext::FIELD_BATCH . '/create';
    return $this->doRequest($url, 'POST', [], $bulk);
  }

  /**
   * Loads the current USER type fields from coveo.
   *
   * 'origin' query parameter: Enum values :"ALL","USER","SYSTEM", Default Value: ALL
   * The origin of the fields to list.
   * - ALL
   * - USER: Created by a Coveo Cloud organization member or by a source addition (each source comes with a set of standard fields).
   * - SYSTEM: Required by the Coveo index and applies to all sources.
   *
   * @return mixed
   *   Response or FALSE.
   *
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/getFieldsUsingGET_2
   */
  public function loadUserFields() {
    $filters = [
      'facet' => "ALL",
      "origin" => "ALL",
      "origin" => "USER",
      "sort" => "ALL",
    ];
    // @todo add enum for origin query parameter
    $url = $this->clientContext->fieldHost . '/indexes/fields/search';
    $user_fields = [];
    $user_fields = $this->doRequest($url, 'GET', $filters, []);
    return $user_fields;
  }

  /**
   * Retrieves a filtered and paginated list of fields from the index of the target Coveo Cloud organization.
   *
   * @param string $name
   *   A substring that must appear in the name property of a field in order for this field to be included in the results.
   * @param string $origin
   *   The origin of the fields to list. "ALL" "SYSTEM" "USER".
   * @param array $options
   *   Array of options for paged coveo request.
   *
   * @return array
   *   Return all the fields.
   *
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/getFieldsUsingPOST_6
   */
  public function listFields(string $name, $origin = 'USER', $options = ['perPage' => '50', 'sortBy' => 'name']) {
    $filters = [
      "name" => $name,
      "origin" => $origin,
    ];

    // @todo add enum for origin query parameter
    $url = $this->clientContext->fieldHost . '/indexes/fields/search';
    $user_fields = [];

    $batch_payload = ["filters" => $filters] + $options;
    $batch = $this->doRequest($url, 'POST', [], $batch_payload);
    if ($batch === FALSE || $batch === NULL) {
      return [];
    }
    $user_fields = array_merge($user_fields, $batch['items']);

    // get additionla pages.
    for ($i = 1; $i < $batch['totalPages']; $i++) {
      $batch_payload['page'] = $i;
      $batch = $this->doRequest($url, 'POST', [], $batch_payload);
      if ($batch !== FALSE || $batch !== NULL) {
        $user_fields = array_merge($user_fields, $batch['items']);
      }
    }
    return $user_fields;
  }

  /**
   * Get field by field_id/coveo name.
   *
   * @return mixed
   *   Response or FALSE.
   *
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/updateFieldsUsingPUT_2
   */
  public function loadField($field_id) {
    if (empty($field_id)) {
      return FALSE;
    }
    $url = $this->clientContext->fieldHost . '/indexes/fields/' . $field_id;
    return $this->doRequest($url, 'GET', [], []);
  }

  /**
   * Get coveo system fields.
   *
   * @return mixed
   *   Response or FALSE.
   *  @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/getFieldsUsingGET_2
   */
  public function loadSystemFields() {
    $url = $this->clientContext->fieldHost . $this->clientContext::FIELD_PAGE;
    return $this->doRequest($url, 'GET', ['origin' => 'SYSTEM'], []);
  }

  /**
   * Updates one or more field in coveo.
   *
   * @param array $bulk
   *   An array to bulk create fields.
   *
   * @return mixed
   *   The response.
   *
   * @throws \Drupal\Exception\
   *
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/updateFieldsUsingPUT_2
   */
  public function updateFields(array $bulk = []) {
    if (empty($bulk)) {
      return FALSE;
    }
    $url = $this->clientContext->fieldHost . $this->clientContext::FIELD_BATCH . '/update';
    return $this->doRequest($url, 'PUT', [], $bulk);
  }

  /**
   * Deletes fields from coveo.
   *
   * @param array $bulk
   *   A bulk group to delete.
   *
   * @return mixed
   *   The response.
   *
   * @throws \Drupal\Exception\
   *
   * @see https://docs.coveo.com/en/8/cloud-v2-api-reference/field-api#operation/removeFieldsUsingDELETE_2
   */
  public function delete(array $bulk = []) {
    if (empty($bulk)) {
      return FALSE;
    }
    $fields = ['fields' => implode(',', $bulk)];
    $url = $this->clientContext->fieldHost . $this->clientContext::FIELD_BATCH . '/delete';
    return $this->doRequest($url, 'DELETE', $fields, []);
  }

  /**
   * Get error message from Coveo response.
   *
   * @param \GuzzleHttp\Exception\RequestException $e
   *   Client Exception.
   */
  public function handleErrorMessage($e) {
    if ($e->hasResponse()) {
      $error_response = $e->getResponse();
      $error_response_body = json_decode($error_response->getBody(), TRUE);
      $this->logException($e);
      \Drupal::messenger()->addError(t('Coveo request failed with code: %error_code and message: %message.', ['%error_code' => $error_response_body['errorCode'], '%message' => $error_response_body['message']]));
    }
    return FALSE;
  }

}
